#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import sys

if sys.version_info < (3, 0):
    raise RuntimeError('ParlAI requires Python 3.')
