#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Core ParlAI library.

This module contains most of the core abstractions and abstract classes used across all
of ParlAI. This includes teachers, worlds, models, general utility functions, and much
more.
"""
